diff -Naur vim72-orig/runtime/doc/if_lua.txt vim72/runtime/doc/if_lua.txt
--- vim72-orig/runtime/doc/if_lua.txt	1969-12-31 19:00:00.000000000 -0500
+++ vim72/runtime/doc/if_lua.txt	2008-08-31 23:34:01.000000000 -0400
@@ -0,0 +1,239 @@
+*if_lua.txt*    For Vim version 7.2.  Last change: 2008 Aug 31
+
+
+		  VIM REFERENCE MANUAL    by Luis Carvalho
+
+
+The Lua Interface to Vim				*lua* *Lua*
+
+1. Commands			|lua-commands|
+2. The vim module		|lua-vim|
+3. Buffer userdata		|lua-buffer|
+4. Window userdata		|lua-window|
+
+{Vi does not have any of these commands}
+
+The Lua interface is available only when Vim was compiled with the
+|+lua| feature.
+
+==============================================================================
+1. Commands						*lua-commands*
+
+							*:lua*
+:[range]lua {chunk}
+			Execute Lua chunk {chunk}.    {not in Vi}
+
+Examples:
+>
+	:lua print("Hello, Vim!")
+	:lua local curbuf = vim.buffer() curbuf[7] = "line #7"
+<
+
+:[range]lua << {endmarker}
+{script}
+{endmarker}
+			Execute Lua script {script}.  {not in Vi}
+			Note: This command doesn't work when the Lua
+			feature wasn't compiled in.  To avoid errors, see
+			|script-here|.
+
+{endmarker} must NOT be preceded by any white space.  If {endmarker} is
+omitted from after the "<<", a dot '.' must be used after {script}, like
+for the |:append| and |:insert| commands.
+This form of the |:lua| command is mainly useful for including Lua code
+in Vim scripts.
+
+Example:
+>
+	function! CurrentLineInfo()
+	lua << EOF
+	local linenr = vim.window().line
+	local curline = vim.buffer()[linenr]
+	print(string.format("Current line [%d] has %d chars",
+		linenr, #curline))
+	EOF
+	endfunction
+<
+
+							*:luado*
+:[range]luado {body}	Execute Lua function$function (line)${body}$end$ for
+			each line in the [range], with the function argument
+			being set to the text of each line in turn, without a
+			trailing <EOL>. If the value returned by the function
+			is a string it becomes the text of the line in the
+			current turn. The default for [range] is the whole
+			file: "1,$".                  {not in Vi}
+
+Examples:
+>
+	:luado return string.format("%s\t%d", line:reverse(), #line)
+
+	:lua require"lpeg"
+	:lua -- balanced parenthesis grammar:
+	:lua bp = lpeg.P{ "(" * ((1 - lpeg.S"()") + lpeg.V(1))^0 * ")" }
+	:luado if bp:match(line) then return "-->\t" .. line end
+<
+
+							*:luafile*
+:[range]luafile {file}
+			Execute Lua script in {file}. {not in Vi}
+			The whole argument is used as a single file name.
+
+Examples:
+>
+	:luafile script.lua
+	:luafile %
+<
+
+All these commands execute a Lua chunk from either the command line (:lua and
+:luado) or a file (:luafile) with the given line [range]. Similarly to the Lua
+interpreter, each chunk has its own scope and so only global variables are
+shared between command calls. Lua default libraries$table$,$string$,$math$,
+and$package$are available,$io$and$debug$are not, and$os$is restricted to
+functions$date$,$clock$,$time$,$difftime$, and$getenv$. In addition,
+Lua$print$function has its output redirected to the Vim message area, with
+arguments separated by a white space instead of a tab.
+
+Lua uses the "vim" module (see|lua-vim|) to issue commands to Vim
+and manage buffers (|lua-buffer|) and windows (|lua-window|). However,
+procedures that alter buffer content, open new buffers, and change cursor
+position are restricted when the command is executed in the|sandbox|.
+
+
+==============================================================================
+2. The vim module					*lua-vim*
+
+Lua interfaces Vim through the "vim" module. The first and last line of the
+input range are stored in$vim.firstline$and$vim.lastline$respectively. The
+module also includes routines for buffer, window, and current line queries,
+Vim evaluation and command execution, and others.
+
+	$vim.isbuffer(value)$	Returns#true#if$value$is a buffer userdata and
+				$false$otherwise (see|lua-buffer|).
+
+	$vim.buffer($[arg]$)$	If$arg$is a number, returns buffer with number
+				$arg$in the buffer list or, if$arg$is
+				a string, returns buffer whose full or short
+				name is$arg$. In both cases, returns#nil#if
+				the buffer is not found. Otherwise, if
+				$toboolean(arg)$is#true#returns the first
+				buffer in the buffer list or else the current
+				buffer.
+
+	$vim.iswindow(value)$	Returns#true#if$value$is a window userdata and
+				$false$otherwise (see|lua-window|).
+
+	$vim.window($[arg]$)$	If$arg$is a number, returns window with number
+				$arg$or#nil#if not found. Otherwise, if
+				$toboolean(arg)$is#true#returns the first
+				window or else the current window.
+
+	$vim.command(${cmd}$)$	Executes the vim (ex-mode) command {cmd}.
+				Examples: >
+					:lua vim.command"set tw=60"
+					:lua vim.command"normal ddp"
+<
+	$vim.eval(${expr}$)$	Evaluates expression {expr} (see|expression|),
+				converts the result to Lua, and returns it.
+				Vim strings and numbers are directly converted
+				to Lua strings and numbers respectively. Vim
+				lists and dictionaries are converted to Lua
+				tables (lists become integer-keyed tables).
+				Examples: >
+					:lua tw = vim.eval"&tw"
+					:lua print(vim.eval"{'a': 'one'}".a)
+<
+	$vim.line()$		Returns the current line (without the trailing
+				<EOL>), a Lua string.
+
+	$vim.beep()$		Beeps.
+
+	$vim.open(${fname}$)$	Opens a new buffer for file {fname} and
+				returns it. Note that the buffer is not set as
+				current.
+
+
+==============================================================================
+3. Buffer userdata					*lua-buffer*
+
+Buffer userdata represent vim buffers. A buffer userdata$b$has the following
+properties and methods:
+
+Properties
+----------
+	#o#$b()$sets$b$as the current buffer.
+	#o#$#b$is the number of lines in buffer$b$.
+	#o#$b[k]$represents line number$k$:$b[k] = newline$replaces line$k$
+	    with string$newline$and$b[k] =$#nil#deletes line$k$.
+	#o#$b.name$contains the short name of buffer$b$(read-only).
+	#o#$b.fname$contains the full name of buffer$b$(read-only).
+	#o#$b.number$contains the position of buffer$b$in the buffer list
+	    (read-only).
+
+Methods
+-------
+	#o#$b:insert(newline$[, pos]$)$inserts string$newline$at position$pos$
+	    in the buffer. The default value for$pos$is$#b + 1$. If$pos == 0$
+	    then$newline$becomes the first line in the buffer.
+	#o#$b:next()$returns the buffer next to$b$in the buffer list.
+	#o#$b:previous()$returns the buffer previous to$b$in the buffer list.
+	#o#$b:isvalid()$returns#true#if buffer$b$corresponds to a "real" (not
+	    freed from memory) Vim buffer.
+
+Examples:
+>
+	:lua b = vim.buffer() -- current buffer
+	:lua print(b.name, b.number)
+	:lua b[1] = "first line"
+	:lua b:insert("FIRST!", 0)
+	:lua b[1] = nil -- delete top line
+	:lua for i=1,3 do b:insert(math.random()) end
+	:3,4lua for i=vim.lastline,vim.firstline,-1 do b[i] = nil end
+	:lua vim.open"myfile"() -- open buffer and set it as current
+
+	function! ListBuffers()
+	lua << EOF
+	local b = vim.buffer(true) -- first buffer in list
+	while b ~= nil do
+		print(b.number, b.name, #b)
+		b = b:next()
+	end
+	vim.beep()
+	EOF
+	endfunction
+<
+
+==============================================================================
+4. Window userdata					*lua-window*
+
+Window objects represent vim windows. A window userdata$w$has the following
+properties and methods:
+
+Properties
+----------
+	#o#$w()$sets$w$as the current window.
+	#o#$w.buffer$contains the buffer of window$w$(read-only).
+	#o#$w.line$represents the cursor line position in window$w$.
+	#o#$w.col$represents the cursor column position in window$w$.
+	#o#$w.width$represents the width of window$w$.
+	#o#$w.height$represents the height of window$w$.
+
+Methods
+-------
+	#o#$w:next()$returns the window next to$w$.
+	#o#$w:previous()$returns the window previous to$w$.
+	#o#$w:isvalid()$returns#true#if window$w$corresponds to a "real" (not
+	    freed from memory) Vim window.
+
+Examples:
+>
+	:lua w = vim.window() -- current window
+	:lua print(w.buffer.name, w.line, w.col)
+	:lua w.width = w.width + math.random(10)
+	:lua w.height = 2 * math.random() * w.height
+	:lua n,w = 0,vim.window(true) while w~=nil do n,w = n + 1,w:next() end
+	:lua print("There are " .. n .. " windows")
+<
+
+==============================================================================
+ vim:tw=78:ts=8:ft=help:norl:
diff -Naur vim72-orig/runtime/doc/tags vim72/runtime/doc/tags
--- vim72-orig/runtime/doc/tags	2008-08-09 10:47:14.000000000 -0400
+++ vim72/runtime/doc/tags	2008-09-01 01:41:36.000000000 -0400
@@ -2318,6 +2318,9 @@
 :lt	tagsrch.txt	/*:lt*
 :ltag	tagsrch.txt	/*:ltag*
 :lu	map.txt	/*:lu*
+:lua	if_lua.txt	/*:lua*
+:luado	if_lua.txt	/*:luado*
+:luafile	if_lua.txt	/*:luafile*
 :lunmap	map.txt	/*:lunmap*
 :lv	quickfix.txt	/*:lv*
 :lvimgrep	quickfix.txt	/*:lvimgrep*
@@ -4189,6 +4192,7 @@
 Linux-backspace	options.txt	/*Linux-backspace*
 List	eval.txt	/*List*
 Lists	eval.txt	/*Lists*
+Lua	if_lua.txt	/*Lua*
 M	motion.txt	/*M*
 MDI	starting.txt	/*MDI*
 MS-DOS	os_msdos.txt	/*MS-DOS*
@@ -6109,6 +6113,7 @@
 idl-syntax	syntax.txt	/*idl-syntax*
 idl.vim	syntax.txt	/*idl.vim*
 if_cscop.txt	if_cscop.txt	/*if_cscop.txt*
+if_lua.txt	if_lua.txt	/*if_lua.txt*
 if_mzsch.txt	if_mzsch.txt	/*if_mzsch.txt*
 if_ole.txt	if_ole.txt	/*if_ole.txt*
 if_perl.txt	if_perl.txt	/*if_perl.txt*
@@ -6279,6 +6284,11 @@
 long-lines	version5.txt	/*long-lines*
 lowercase	change.txt	/*lowercase*
 lpc.vim	syntax.txt	/*lpc.vim*
+lua	if_lua.txt	/*lua*
+lua-buffer	if_lua.txt	/*lua-buffer*
+lua-commands	if_lua.txt	/*lua-commands*
+lua-vim	if_lua.txt	/*lua-vim*
+lua-window	if_lua.txt	/*lua-window*
 lua.vim	syntax.txt	/*lua.vim*
 m	motion.txt	/*m*
 m'	motion.txt	/*m'*
diff -Naur vim72-orig/src/buffer.c vim72/src/buffer.c
--- vim72-orig/src/buffer.c	2008-08-06 07:00:48.000000000 -0400
+++ vim72/src/buffer.c	2008-08-30 10:32:59.000000000 -0400
@@ -598,6 +598,9 @@
     buf_T	*buf;
 {
     free_buffer_stuff(buf, TRUE);
+#ifdef FEAT_LUA
+    lua_buffer_free(buf);
+#endif
 #ifdef FEAT_MZSCHEME
     mzscheme_buffer_free(buf);
 #endif
diff -Naur vim72-orig/src/config.aap.in vim72/src/config.aap.in
--- vim72-orig/src/config.aap.in	2008-07-13 15:21:00.000000000 -0400
+++ vim72/src/config.aap.in	2008-08-30 18:35:48.000000000 -0400
@@ -22,6 +22,12 @@
 X_EXTRA_LIBS	= @X_EXTRA_LIBS@
 X_LIBS		= @X_LIB@
 
+LUA_LIBS	= @LUA_LIBS@
+LUA_SRC	= @LUA_SRC@
+LUA_OBJ	= @LUA_OBJ@
+LUA_CFLAGS	= @LUA_CFLAGS@
+LUA_PRO	= @LUA_PRO@
+
 MZSCHEME_LIBS	= @MZSCHEME_LIBS@
 MZSCHEME_SRC	= @MZSCHEME_SRC@
 MZSCHEME_OBJ	= @MZSCHEME_OBJ@
diff -Naur vim72-orig/src/config.h.in vim72/src/config.h.in
--- vim72-orig/src/config.h.in	2008-06-21 11:01:41.000000000 -0400
+++ vim72/src/config.h.in	2008-08-30 17:03:11.000000000 -0400
@@ -284,6 +284,9 @@
 /* Define if you want huge features. */
 #undef FEAT_HUGE
 
+/* Define if you want to include the Lua interpreter. */
+#undef FEAT_LUA
+
 /* Define if you want to include the MzScheme interpreter. */
 #undef FEAT_MZSCHEME
 
diff -Naur vim72-orig/src/config.mk.in vim72/src/config.mk.in
--- vim72-orig/src/config.mk.in	2008-06-21 11:56:41.000000000 -0400
+++ vim72/src/config.mk.in	2008-08-30 18:28:37.000000000 -0400
@@ -36,6 +36,12 @@
 X_EXTRA_LIBS	= @X_EXTRA_LIBS@
 X_LIBS		= @X_LIB@
 
+LUA_LIBS	= @LUA_LIBS@
+LUA_SRC		= @LUA_SRC@
+LUA_OBJ		= @LUA_OBJ@
+LUA_CFLAGS	= @LUA_CFLAGS@
+LUA_PRO		= @LUA_PRO@
+
 MZSCHEME_LIBS	= @MZSCHEME_LIBS@
 MZSCHEME_SRC	= @MZSCHEME_SRC@
 MZSCHEME_OBJ	= @MZSCHEME_OBJ@
diff -Naur vim72-orig/src/configure.in vim72/src/configure.in
--- vim72-orig/src/configure.in	2008-07-24 08:40:26.000000000 -0400
+++ vim72/src/configure.in	2008-09-01 12:42:16.000000000 -0400
@@ -389,6 +389,105 @@
   AC_MSG_RESULT(yes)
 fi
 
+dnl Check for Lua feature.
+AC_MSG_CHECKING(--enable-luainterp argument)
+AC_ARG_ENABLE(luainterp,
+	[  --enable-luainterp      Include Lua interpreter.], ,
+	[enable_luainterp="no"])
+AC_MSG_RESULT($enable_luainterp)
+
+if test "$enable_luainterp" = "yes"; then
+  dnl -- find the lua executable
+  AC_SUBST(vi_cv_path_lua)
+
+  AC_MSG_CHECKING(--with-lua-prefix argument)
+  AC_ARG_WITH(lua_prefix,
+	[  --with-lua-prefix=PFX   Prefix where Lua is installed.],
+	with_lua_prefix="$withval"; AC_MSG_RESULT($with_lua_prefix),
+	with_lua_prefix="";AC_MSG_RESULT("no"))
+
+  if test "X$with_lua_prefix" != "X"; then
+       vi_cv_path_lua_pfx="$with_lua_prefix"
+  else
+    AC_MSG_CHECKING(LUA_PREFIX environment var)
+    if test "X$LUA_PREFIX" != "X"; then
+	AC_MSG_RESULT("$LUA_PREFIX")
+	vi_cv_path_lua_pfx="$LUA_PREFIX"
+    else
+	AC_MSG_RESULT("not set")
+    fi
+  fi
+
+  LUA_INC=
+  if test "X$vi_cv_path_lua_pfx" != "X"; then
+    AC_MSG_CHECKING(if lua.h can be found in $vi_cv_path_lua_pfx/include)
+    if test -f $vi_cv_path_lua_pfx/include/lua.h; then
+      AC_MSG_RESULT("yes")
+    else
+      AC_MSG_RESULT("no")
+      dnl -- try to find Lua executable
+      AC_PATH_PROG(vi_cv_path_lua, lua)
+      if test "X$vi_cv_path_lua" != "X"; then
+        dnl -- find Lua version
+        AC_CACHE_CHECK(Lua version, vi_cv_version_lua,
+        [ vi_cv_version_lua=`${vi_cv_path_lua} -e "print(_VERSION:sub(5,7))"` ])
+        AC_MSG_CHECKING(if lua.h can be found in $vi_cv_path_lua_pfx/include/lua$vi_cv_version_lua)
+        if test -f $vi_cv_path_lua_pfx/include/lua$vi_cv_version_lua/lua.h; then
+          AC_MSG_RESULT("yes")
+          LUA_INC=/lua$vi_cv_version_lua
+        else
+          AC_MSG_RESULT("no")
+          vi_cv_path_lua_pfx=
+        fi
+      fi
+    fi
+  fi
+
+  if test "X$vi_cv_path_lua_pfx" != "X"; then
+    if test "X$vi_cv_version_lua" != "X"; then
+      dnl Test alternate location using version
+      if test -f "${vi_cv_path_lua_pfx}/lib/liblua$vi_cv_version_lua.a"; then
+        LUA_LIBS="${vi_cv_path_lua_pfx}/lib/liblua$vi_cv_version_lua.a"
+      else
+        LUA_LIBS="-L${vi_cv_path_lua_pfx}/lib -llua$vi_cv_version_lua"
+        if test "$GCC" = yes; then
+          dnl Make Vim remember the path to the library.  For when it's not in
+          dnl $LD_LIBRARY_PATH.
+          LUA_LIBS="$LUA_LIBS -Wl,-rpath -Wl,${vi_cv_path_lua_pfx}/lib"
+        elif test "`(uname) 2>/dev/null`" = SunOS &&
+          uname -r | grep '^5' >/dev/null; then
+          LUA_LIBS="$LUA_LIBS -R ${vi_cv_path_lua_pfx}/lib"
+        fi
+      fi
+    else
+      if test -f "${vi_cv_path_lua_pfx}/lib/liblua.a"; then
+        LUA_LIBS="${vi_cv_path_lua_pfx}/lib/liblua.a"
+      else
+        LUA_LIBS="-L${vi_cv_path_lua_pfx}/lib -llua"
+        if test "$GCC" = yes; then
+          dnl Make Vim remember the path to the library.  For when it's not in
+          dnl $LD_LIBRARY_PATH.
+          LUA_LIBS="$LUA_LIBS -Wl,-rpath -Wl,${vi_cv_path_lua_pfx}/lib"
+        elif test "`(uname) 2>/dev/null`" = SunOS &&
+          uname -r | grep '^5' >/dev/null; then
+          LUA_LIBS="$LUA_LIBS -R ${vi_cv_path_lua_pfx}/lib"
+        fi
+      fi
+    fi
+    LUA_CFLAGS="-I${vi_cv_path_lua_pfx}/include${LUA_INC}"
+    LUA_SRC="if_lua.c"
+    LUA_OBJ="objects/if_lua.o"
+    LUA_PRO="if_lua.pro"
+    AC_DEFINE(FEAT_LUA)
+  fi
+  AC_SUBST(LUA_SRC)
+  AC_SUBST(LUA_OBJ)
+  AC_SUBST(LUA_PRO)
+  AC_SUBST(LUA_LIBS)
+  AC_SUBST(LUA_CFLAGS)
+fi
+
+
 dnl Check for MzScheme feature.
 AC_MSG_CHECKING(--enable-mzschemeinterp argument)
 AC_ARG_ENABLE(mzschemeinterp,
diff -Naur vim72-orig/src/eval.c vim72/src/eval.c
--- vim72-orig/src/eval.c	2008-08-07 15:37:22.000000000 -0400
+++ vim72/src/eval.c	2008-08-30 10:37:42.000000000 -0400
@@ -5835,7 +5835,7 @@
     return item1 == NULL && item2 == NULL;
 }
 
-#if defined(FEAT_PYTHON) || defined(PROTO)
+#if defined(FEAT_PYTHON) || defined(FEAT_LUA) || defined(PROTO)
 /*
  * Return the dictitem that an entry in a hashtable points to.
  */
@@ -11549,6 +11549,11 @@
 #ifdef FEAT_LOCALMAP
 	"localmap",
 #endif
+#ifdef FEAT_LUA
+# ifndef DYNAMIC_LUA
+	"lua",
+# endif
+#endif
 #ifdef FEAT_MENU
 	"menu",
 #endif
@@ -11794,6 +11799,10 @@
 	else if (STRICMP(name, "iconv") == 0)
 	    n = iconv_enabled(FALSE);
 #endif
+#ifdef DYNAMIC_LUA
+	else if (STRICMP(name, "lua") == 0)
+	    n = lua_enabled(FALSE); /* TODO */
+#endif
 #ifdef DYNAMIC_MZSCHEME
 	else if (STRICMP(name, "mzscheme") == 0)
 	    n = mzscheme_enabled(FALSE);
diff -Naur vim72-orig/src/ex_cmds.h vim72/src/ex_cmds.h
--- vim72-orig/src/ex_cmds.h	2008-06-21 14:47:57.000000000 -0400
+++ vim72/src/ex_cmds.h	2008-08-30 23:11:55.000000000 -0400
@@ -571,6 +571,12 @@
 			RANGE|NOTADR|COUNT|TRLBAR|BANG),
 EX(CMD_ltag,		"ltag",	ex_tag,
 			NOTADR|TRLBAR|BANG|WORD1),
+EX(CMD_lua,		"lua",		ex_lua,
+			RANGE|EXTRA|NEEDARG|CMDWIN),
+EX(CMD_luado,		"luado",	ex_luado,
+			RANGE|DFLALL|EXTRA|NEEDARG|CMDWIN),
+EX(CMD_luafile,		"luafile",	ex_luafile,
+			RANGE|FILE1|NEEDARG|CMDWIN),
 EX(CMD_lunmap,		"lunmap",	ex_unmap,
 			EXTRA|TRLBAR|NOTRLCOM|USECTRLV|CMDWIN),
 EX(CMD_lvimgrep,	"lvimgrep",	ex_vimgrep,
diff -Naur vim72-orig/src/ex_docmd.c vim72/src/ex_docmd.c
--- vim72-orig/src/ex_docmd.c	2008-07-26 07:51:05.000000000 -0400
+++ vim72/src/ex_docmd.c	2008-08-30 23:13:01.000000000 -0400
@@ -241,6 +241,11 @@
 # define ex_spellinfo		ex_ni
 # define ex_spellrepall		ex_ni
 #endif
+#ifndef FEAT_LUA
+# define ex_lua			ex_script_ni
+# define ex_luado		ex_ni
+# define ex_luafile		ex_ni
+#endif
 #ifndef FEAT_MZSCHEME
 # define ex_mzscheme		ex_script_ni
 # define ex_mzfile		ex_ni
@@ -2524,6 +2529,7 @@
 	    case CMD_leftabove:
 	    case CMD_let:
 	    case CMD_lockmarks:
+	    case CMD_lua:
 	    case CMD_match:
 	    case CMD_mzscheme:
 	    case CMD_perl:
diff -Naur vim72-orig/src/feature.h vim72/src/feature.h
--- vim72-orig/src/feature.h	2008-08-06 07:00:39.000000000 -0400
+++ vim72/src/feature.h	2008-08-30 10:43:38.000000000 -0400
@@ -1162,6 +1162,7 @@
 /*
  * These features can only be included by using a configure argument.  See the
  * Makefile for a line to uncomment.
+ * +lua                 Lua interface: "--enable-luainterp"
  * +mzscheme		MzScheme interface: "--enable-mzscheme"
  * +perl		Perl interface: "--enable-perlinterp"
  * +python		Python interface: "--enable-pythoninterp"
diff -Naur vim72-orig/src/if_lua.c vim72/src/if_lua.c
--- vim72-orig/src/if_lua.c	1969-12-31 19:00:00.000000000 -0500
+++ vim72/src/if_lua.c	2008-08-31 22:50:25.000000000 -0400
@@ -0,0 +1,807 @@
+/* vi:set ts=8 sts=4 sw=4:
+ *
+ * VIM - Vi IMproved	by Bram Moolenaar
+ *
+ * Lua interface by Luis Carvalho
+ *
+ * Do ":help uganda"  in Vim to read copying and usage conditions.
+ * Do ":help credits" in Vim to see a list of people who contributed.
+ * See README.txt for an overview of the Vim source code.
+ */
+
+#include <stdio.h>
+#include <string.h>
+#include <lua.h>
+#include <lualib.h>
+#include <lauxlib.h>
+#include "vim.h"
+
+/* Only do the following when the feature is enabled.  Needed for "make
+ * depend". */
+#if defined(FEAT_LUA) || defined(PROTO)
+
+#define LUAVIM_CHUNKNAME "vim chunk"
+#define LUAVIM_NAME "vim"
+
+typedef buf_T *luaV_Buffer;
+typedef win_T *luaV_Window;
+
+static const char LUAVIM_BUFFER[] = "buffer";
+static const char LUAVIM_WINDOW[] = "window";
+static const char LUAVIM_FREE[] = "luaV_free";
+
+#define luaV_getfield(L, s) \
+    lua_pushlightuserdata((L), (void *)(s)); \
+    lua_rawget((L), LUA_REGISTRYINDEX)
+#define luaV_checksandbox(L) \
+    if (sandbox) luaL_error((L), "not allowed in sandbox")
+
+/* =======   Internal   ======= */
+
+static void luaV_newmetatable (lua_State *L, const char *tname) {
+    lua_newtable(L);
+    lua_pushlightuserdata(L, (void *) tname);
+    lua_pushvalue(L, -2);
+    lua_rawset(L, LUA_REGISTRYINDEX);
+}
+
+static void *luaV_toudata (lua_State *L, int ud, const char *tname) {
+    void *p = lua_touserdata(L, ud);
+    if (p != NULL) { /* value is userdata? */
+        if (lua_getmetatable(L, ud)) { /* does it have a metatable? */
+            luaV_getfield(L, tname); /* get metatable */
+            if (lua_rawequal(L, -1, -2)) { /* MTs match? */
+                lua_pop(L, 2); /* MTs */
+                return p;
+            }
+        }
+    }
+    return NULL;
+}
+
+static void *luaV_checkudata (lua_State *L, int ud, const char *tname) {
+    void *p = luaV_toudata(L, ud, tname);
+    if (p == NULL) luaL_typerror(L, ud, tname);
+    return p;
+}
+
+static void luaV_pushtypval (lua_State *L, typval_T *tv) {
+    if (tv == NULL) luaL_error(L, "null type");
+    switch (tv->v_type) {
+        case VAR_STRING:
+            lua_pushstring(L, (char *) tv->vval.v_string);
+            break;
+        case VAR_NUMBER:
+            lua_pushinteger(L, (int) tv->vval.v_number);
+            break;
+#ifdef FEAT_FLOAT
+        case VAR_FLOAT:
+            lua_pushnumber(L, (lua_Number) tv->vval.v_float);
+            break;
+#endif
+        case VAR_LIST: {
+            list_T *l = tv->vval.v_list;
+            if (l != NULL) {
+                /* check cache */
+                lua_pushlightuserdata(L, (void *) l);
+                lua_rawget(L, LUA_ENVIRONINDEX);
+                if (lua_isnil(L, -1)) { /* not interned? */
+                    listitem_T *li;
+                    int n = 0;
+                    lua_newtable(L);
+                    for (li = l->lv_first; li != NULL; li = li->li_next) {
+                        luaV_pushtypval(L, &li->li_tv);
+                        lua_rawseti(L, -2, ++n);
+                    }
+                    lua_pushlightuserdata(L, (void *) l);
+                    lua_pushvalue(L, -2);
+                    lua_rawset(L, LUA_ENVIRONINDEX);
+                }
+            }
+            else lua_pushnil(L);
+            break;
+        }
+        case VAR_DICT: {
+            dict_T *d = tv->vval.v_dict;
+            if (d != NULL) {
+                /* check cache */
+                lua_pushlightuserdata(L, (void *) d);
+                lua_rawget(L, LUA_ENVIRONINDEX);
+                if (lua_isnil(L, -1)) { /* not interned? */
+                    hashtab_T *ht = &d->dv_hashtab;
+                    hashitem_T *hi;
+                    int n = ht->ht_used; /* remaining items */
+                    lua_newtable(L);
+                    for (hi = ht->ht_array; n > 0; hi++) {
+                        if (!HASHITEM_EMPTY(hi)) {
+                            dictitem_T *di = dict_lookup(hi);
+                            luaV_pushtypval(L, &di->di_tv);
+                            lua_setfield(L, -2, (char *) hi->hi_key);
+                            n--;
+                        }
+                    }
+                    lua_pushlightuserdata(L, (void *) d);
+                    lua_pushvalue(L, -2);
+                    lua_rawset(L, LUA_ENVIRONINDEX);
+                }
+            }
+            else lua_pushnil(L);
+            break;
+        }
+        default:
+            luaL_error(L, "invalid type");
+    }
+}
+
+/* =======   Buffer type   ======= */
+
+static luaV_Buffer *luaV_newbuffer (lua_State *L, buf_T *buf) {
+    luaV_Buffer *b = (luaV_Buffer *) lua_newuserdata(L, sizeof(luaV_Buffer));
+    *b = buf;
+    lua_pushlightuserdata(L, (void *) buf);
+    lua_pushvalue(L, -2);
+    lua_rawset(L, LUA_ENVIRONINDEX); /* env[buf] = udata */
+    /* to avoid GC, store as key in env */
+    lua_pushvalue(L, -1);
+    lua_pushboolean(L, 1);
+    lua_rawset(L, LUA_ENVIRONINDEX); /* env[udata] = true */
+    /* set metatable */
+    luaV_getfield(L, LUAVIM_BUFFER);
+    lua_setmetatable(L, -2);
+    return b;
+}
+
+static luaV_Buffer *luaV_pushbuffer (lua_State *L, buf_T *buf) {
+    luaV_Buffer *b = NULL;
+    if (buf == NULL)
+        lua_pushnil(L);
+    else {
+        lua_pushlightuserdata(L, (void *) buf);
+        lua_rawget(L, LUA_ENVIRONINDEX);
+        if (lua_isnil(L, -1)) { /* not interned? */
+            lua_pop(L, 1);
+            b = luaV_newbuffer(L, buf);
+        }
+        else b = (luaV_Buffer *) lua_touserdata(L, -1);
+    }
+    return b;
+}
+
+/* Buffer metamethods */
+
+static int luaV_buffer_tostring (lua_State *L) {
+    lua_pushfstring(L, "%s: %p", LUAVIM_BUFFER, lua_touserdata(L, 1));
+    return 1;
+}
+
+static int luaV_buffer_len (lua_State *L) {
+    luaV_Buffer *b = lua_touserdata(L, 1);
+    lua_pushinteger(L, (*b)->b_ml.ml_line_count);
+    return 1;
+}
+
+static int luaV_buffer_call (lua_State *L) {
+    luaV_Buffer *b = (luaV_Buffer *) lua_touserdata(L, 1);
+    lua_settop(L, 1);
+    set_curbuf(*b, DOBUF_SPLIT);
+    return 1;
+}
+
+static int luaV_buffer_index (lua_State *L) {
+    luaV_Buffer *b = (luaV_Buffer *) lua_touserdata(L, 1);
+    linenr_T n = (linenr_T) lua_tointeger(L, 2);
+    if (n > 0 && n <= (*b)->b_ml.ml_line_count)
+        lua_pushstring(L, (char *) ml_get_buf(*b, n, FALSE));
+    else if (lua_isstring(L, 2)) {
+        const char *s = lua_tostring(L, 2);
+        if (strncmp(s, "name", 4) == 0)
+            lua_pushstring(L, (char *) (*b)->b_sfname);
+        else if (strncmp(s, "fname", 5) == 0)
+            lua_pushstring(L, (char *) (*b)->b_ffname);
+        else if (strncmp(s, "number", 6) == 0)
+            lua_pushinteger(L, (*b)->b_fnum);
+        /* methods */
+        else if (strncmp(s,   "insert", 6) == 0
+                || strncmp(s, "next", 4) == 0
+                || strncmp(s, "previous", 8) == 0
+                || strncmp(s, "isvalid", 7) == 0) {
+            lua_getmetatable(L, 1);
+            lua_getfield(L, -1, s);
+        }
+        else lua_pushnil(L);
+    }
+    else lua_pushnil(L);
+    return 1;
+}
+
+static int luaV_buffer_newindex (lua_State *L) {
+    luaV_Buffer *b = (luaV_Buffer *) lua_touserdata(L, 1);
+    linenr_T n = (linenr_T) luaL_checkinteger(L, 2);
+#ifdef HAVE_SANDBOX
+    luaV_checksandbox(L);
+#endif
+    if (n < 1 || n > (*b)->b_ml.ml_line_count)
+        luaL_error(L, "invalid line number");
+    if (lua_isnil(L, 3)) { /* delete line */
+        buf_T *buf = curbuf;
+        curbuf = *b;
+        if (u_savedel(n, 1L) == FAIL) {
+            curbuf = buf;
+            luaL_error(L, "cannot save undo information");
+        }
+        else if (ml_delete(n, FALSE) == FAIL) {
+            curbuf = buf;
+            luaL_error(L, "cannot delete line");
+        }
+        else {
+            deleted_lines_mark(n, 1L);
+            if (*b == curwin->w_buffer) { /* fix cursor in current window? */
+                if (curwin->w_cursor.lnum >= n) {
+                    if (curwin->w_cursor.lnum > n) {
+                        curwin->w_cursor.lnum -= 1;
+                        check_cursor_col();
+                    }
+                    else check_cursor();
+                    changed_cline_bef_curs();
+                }
+                invalidate_botline();
+            }
+        }
+        curbuf = buf;
+    }
+    else if (lua_isstring(L, 3)) { /* update line */
+        char_u *s = (char_u *) lua_tostring(L, 3);
+        buf_T *buf = curbuf;
+        curbuf = *b;
+        if (u_savesub(n) == FAIL) {
+            curbuf = buf;
+            luaL_error(L, "cannot save undo information");
+        }
+        else if (ml_replace(n, s, TRUE) == FAIL) {
+            curbuf = buf;
+            luaL_error(L, "cannot replace line");
+        }
+        else changed_bytes(n, 0);
+        curbuf = buf;
+        if (*b == curwin->w_buffer)
+            check_cursor_col();
+    }
+    else
+        luaL_error(L, "wrong argument to change line");
+    return 0;
+}
+
+static int luaV_buffer_insert (lua_State *L) {
+    luaV_Buffer *b = luaV_checkudata(L, 1, LUAVIM_BUFFER);
+    char_u *s = (char_u *) luaL_checkstring(L, 2);
+    linenr_T last = (*b)->b_ml.ml_line_count;
+    linenr_T n = (linenr_T) luaL_optinteger(L, 3, last);
+    buf_T *buf;
+#ifdef HAVE_SANDBOX
+    luaV_checksandbox(L);
+#endif
+    /* fix insertion line */
+    if (n < 0) n = 0;
+    if (n > last) n = last;
+    /* insert */
+    buf = curbuf;
+    curbuf = *b;
+    if (u_save(n, n + 1) == FAIL) {
+        curbuf = buf;
+        luaL_error(L, "cannot save undo information");
+    }
+    else if (ml_append(n, s, 0, FALSE) == FAIL) {
+        curbuf = buf;
+        luaL_error(L, "cannot insert line");
+    }
+    else
+        appended_lines_mark(n, 1L);
+    curbuf = buf;
+    update_screen(VALID);
+    return 0;
+}
+
+static int luaV_buffer_next (lua_State *L) {
+    luaV_Buffer *b = luaV_checkudata(L, 1, LUAVIM_BUFFER);
+    luaV_pushbuffer(L, (*b)->b_next);
+    return 1;
+}
+
+static int luaV_buffer_previous (lua_State *L) {
+    luaV_Buffer *b = luaV_checkudata(L, 1, LUAVIM_BUFFER);
+    luaV_pushbuffer(L, (*b)->b_prev);
+    return 1;
+}
+
+static int luaV_buffer_isvalid (lua_State *L) {
+    luaV_Buffer *b = luaV_checkudata(L, 1, LUAVIM_BUFFER);
+    lua_pushlightuserdata(L, (void *) (*b));
+    lua_rawget(L, LUA_ENVIRONINDEX);
+    lua_pushboolean(L, !lua_isnil(L, -1));
+    return 1;
+}
+
+static const luaL_Reg luaV_Buffer_mt[] = {
+    {"__tostring", luaV_buffer_tostring},
+    {"__len", luaV_buffer_len},
+    {"__call", luaV_buffer_call},
+    {"__index", luaV_buffer_index},
+    {"__newindex", luaV_buffer_newindex},
+    {"insert", luaV_buffer_insert},
+    {"next", luaV_buffer_next},
+    {"previous", luaV_buffer_previous},
+    {"isvalid", luaV_buffer_isvalid},
+    {NULL, NULL}
+};
+
+
+/* =======   Window type   ======= */
+
+static luaV_Window *luaV_newwindow (lua_State *L, win_T *win) {
+    luaV_Window *w = (luaV_Window *) lua_newuserdata(L, sizeof(luaV_Window));
+    *w = win;
+    lua_pushlightuserdata(L, (void *) win);
+    lua_pushvalue(L, -2);
+    lua_rawset(L, LUA_ENVIRONINDEX); /* env[win] = udata */
+    /* to avoid GC, store as key in env */
+    lua_pushvalue(L, -1);
+    lua_pushboolean(L, 1);
+    lua_rawset(L, LUA_ENVIRONINDEX); /* env[udata] = true */
+    /* set metatable */
+    luaV_getfield(L, LUAVIM_WINDOW);
+    lua_setmetatable(L, -2);
+    return w;
+}
+
+static luaV_Window *luaV_pushwindow (lua_State *L, win_T *win) {
+    luaV_Window *w = NULL;
+    if (win == NULL)
+        lua_pushnil(L);
+    else {
+        lua_pushlightuserdata(L, (void *) win);
+        lua_rawget(L, LUA_ENVIRONINDEX);
+        if (lua_isnil(L, -1)) { /* not interned? */
+            lua_pop(L, 1);
+            w = luaV_newwindow(L, win);
+        }
+        else w = (luaV_Window *) lua_touserdata(L, -1);
+    }
+    return w;
+}
+
+/* Window metamethods */
+
+static int luaV_window_tostring (lua_State *L) {
+    lua_pushfstring(L, "%s: %p", LUAVIM_WINDOW, lua_touserdata(L, 1));
+    return 1;
+}
+
+static int luaV_window_call (lua_State *L) {
+    luaV_Window *w = (luaV_Window *) lua_touserdata(L, 1);
+    lua_settop(L, 1);
+    win_goto(*w);
+    return 1;
+}
+
+static int luaV_window_index (lua_State *L) {
+    luaV_Window *w = (luaV_Window *) lua_touserdata(L, 1);
+    const char *s = luaL_checkstring(L, 2);
+    if (strncmp(s, "buffer", 6) == 0)
+        luaV_pushbuffer(L, (*w)->w_buffer);
+    else if (strncmp(s, "line", 4) == 0)
+        lua_pushinteger(L, (*w)->w_cursor.lnum);
+    else if (strncmp(s, "col", 3) == 0)
+        lua_pushinteger(L, (*w)->w_cursor.col + 1);
+#ifdef FEAT_VERTSPLIT
+    else if (strncmp(s, "width", 5) == 0)
+        lua_pushinteger(L, W_WIDTH((*w)));
+#endif
+    else if (strncmp(s, "height", 6) == 0)
+        lua_pushinteger(L, (*w)->w_height);
+    /* methods */
+    else if (strncmp(s,   "next", 4) == 0
+            || strncmp(s, "previous", 8) == 0
+            || strncmp(s, "isvalid", 7) == 0) {
+        lua_getmetatable(L, 1);
+        lua_getfield(L, -1, s);
+    }
+    else
+        lua_pushnil(L);
+    return 1;
+}
+
+static int luaV_window_newindex (lua_State *L) {
+    luaV_Window *w = (luaV_Window *) lua_touserdata(L, 1);
+    const char *s = luaL_checkstring(L, 2);
+    int v = luaL_checkinteger(L, 3);
+    if (strncmp(s, "line", 4) == 0) {
+#ifdef HAVE_SANDBOX
+        luaV_checksandbox(L);
+#endif
+        if (v < 1 || v > (*w)->w_buffer->b_ml.ml_line_count)
+            luaL_error(L, "line out of range");
+        (*w)->w_cursor.lnum = v;
+        update_screen(VALID);
+    }
+    else if (strncmp(s, "col", 3) == 0) {
+#ifdef HAVE_SANDBOX
+        luaV_checksandbox(L);
+#endif
+        (*w)->w_cursor.col = v - 1;
+        update_screen(VALID);
+    }
+#ifdef FEAT_VERTSPLIT
+    else if (strncmp(s, "width", 5) == 0) {
+        win_T *win = curwin;
+#ifdef FEAT_GUI
+        need_mouse_correct = TRUE;
+#endif
+        curwin = *w;
+        win_setwidth(v);
+        curwin = win;
+    }
+#endif
+    else if (strncmp(s, "height", 6) == 0) {
+        win_T *win = curwin;
+#ifdef FEAT_GUI
+        need_mouse_correct = TRUE;
+#endif
+        curwin = *w;
+        win_setheight(v);
+        curwin = win;
+    }
+    else
+        luaL_error(L, "invalid window property: `%s'", s);
+    return 0;
+}
+
+static int luaV_window_next (lua_State *L) {
+    luaV_Window *w = luaV_checkudata(L, 1, LUAVIM_WINDOW);
+    luaV_pushwindow(L, (*w)->w_next);
+    return 1;
+}
+
+static int luaV_window_previous (lua_State *L) {
+    luaV_Window *w = luaV_checkudata(L, 1, LUAVIM_WINDOW);
+    luaV_pushwindow(L, (*w)->w_prev);
+    return 1;
+}
+
+static int luaV_window_isvalid (lua_State *L) {
+    luaV_Window *w = luaV_checkudata(L, 1, LUAVIM_WINDOW);
+    lua_pushlightuserdata(L, (void *) (*w));
+    lua_rawget(L, LUA_ENVIRONINDEX);
+    lua_pushboolean(L, !lua_isnil(L, -1));
+    return 1;
+}
+
+static const luaL_Reg luaV_Window_mt[] = {
+    {"__tostring", luaV_window_tostring},
+    {"__call", luaV_window_call},
+    {"__index", luaV_window_index},
+    {"__newindex", luaV_window_newindex},
+    {"next", luaV_window_next},
+    {"previous", luaV_window_previous},
+    {"isvalid", luaV_window_isvalid},
+    {NULL, NULL}
+};
+
+
+/* =======   Vim module   ======= */
+
+static int luaV_print (lua_State *L) {
+    int i, n = lua_gettop(L); /* nargs */
+    luaL_Buffer b;
+    luaL_buffinit(L, &b);
+    lua_getglobal(L, "tostring");
+    for (i = 1; i <= n; i++) {
+        const char *s;
+        size_t l;
+        lua_pushvalue(L, -1); /* tostring */
+        lua_pushvalue(L, i); /* arg */
+        lua_call(L, 1, 1);
+        s = lua_tolstring(L, -1, &l);
+        if (s == NULL)
+            return luaL_error(L, "cannot convert to string");
+        if (i > 1) luaL_addchar(&b, ' '); /* use space instead of tab */
+        luaL_addlstring(&b, s, l);
+        lua_pop(L, 1);
+    }
+    luaL_pushresult(&b);
+    MSG(lua_tostring(L, -1)); /* Vim message */
+    lua_pop(L, 1);
+    return 0;
+}
+
+static int luaV_command (lua_State *L) {
+    do_cmdline_cmd((char_u *) luaL_checkstring(L, 1));
+    update_screen(VALID);
+    return 0;
+}
+
+static int luaV_eval (lua_State *L) {
+    typval_T *tv = eval_expr((char_u *) luaL_checkstring(L, 1), NULL);
+    if (tv == NULL) luaL_error(L, "invalid expression");
+    luaV_pushtypval(L, tv);
+    return 1;
+}
+
+static int luaV_beep (lua_State *L) {
+    vim_beep();
+    return 0;
+}
+
+static int luaV_line (lua_State *L) {
+    lua_pushstring(L, (char *) ml_get(curwin->w_cursor.lnum));
+    return 1;
+}
+
+static int luaV_buffer (lua_State *L) {
+    buf_T *buf;
+    if (lua_isstring(L, 1)) { /* get by number or name? */
+        if (lua_isnumber(L, 1)) { /* by number? */
+            int n = lua_tonumber(L, 1);
+            for (buf = firstbuf; buf != NULL; buf = buf->b_next)
+                if (buf->b_fnum == n) break;
+        }
+        else { /* by name */
+            size_t l;
+            const char *s = lua_tolstring(L, 1, &l);
+            for (buf = firstbuf; buf != NULL; buf = buf->b_next) {
+                if (buf->b_ffname == NULL || buf->b_sfname == NULL) {
+                    if (l == 0) break;
+                }
+                else if (strncmp(s, buf->b_ffname, l) == 0
+                        || strncmp(s, buf->b_sfname, l) == 0)
+                    break;
+            }
+        }
+        if (buf == NULL) /* not found? */
+            lua_pushnil(L);
+        else
+            luaV_pushbuffer(L, buf);
+    }
+    else {
+        buf = (lua_toboolean(L, 1)) ? firstbuf : curbuf; /* first buffer? */
+        luaV_pushbuffer(L, buf);
+    }
+    return 1;
+}
+
+static int luaV_window (lua_State *L) {
+    win_T *win;
+    if (lua_isnumber(L, 1)) { /* get by number? */
+        int n = lua_tonumber(L, 1);
+        for (win = firstwin; win != NULL; win = win->w_next, n--)
+            if (n == 1) break;
+        if (win == NULL) /* not found? */
+            lua_pushnil(L);
+        else
+            luaV_pushwindow(L, win);
+    }
+    else {
+        win = (lua_toboolean(L, 1)) ? firstwin : curwin; /* first window? */
+        luaV_pushwindow(L, win);
+    }
+    return 1;
+}
+
+static int luaV_open (lua_State *L) {
+    luaV_Buffer *b;
+    char_u *s = NULL;
+#ifdef HAVE_SANDBOX
+    luaV_checksandbox(L);
+#endif
+    if (lua_isstring(L, 1)) s = (char_u *) lua_tostring(L, 1);
+    b = luaV_pushbuffer(L, buflist_new(s, NULL, 1L, BLN_LISTED));
+    return 1;
+}
+
+static int luaV_isbuffer (lua_State *L) {
+    lua_pushboolean(L, luaV_toudata(L, 1, LUAVIM_BUFFER) != NULL);
+    return 1;
+}
+
+static int luaV_iswindow (lua_State *L) {
+    lua_pushboolean(L, luaV_toudata(L, 1, LUAVIM_WINDOW) != NULL);
+    return 1;
+}
+
+/* for freeing buffer and window objects; lightuserdata as arg */
+static luaV_free (lua_State *L) {
+    lua_pushvalue(L, 1); /* lightudata */
+    lua_rawget(L, LUA_ENVIRONINDEX);
+    if (!lua_isnil(L, -1)) {
+        lua_pushnil(L);
+        lua_rawset(L, LUA_ENVIRONINDEX); /* env[udata] = nil */
+        lua_pushnil(L);
+        lua_rawset(L, LUA_ENVIRONINDEX); /* env[lightudata] = nil */
+    }
+    return 0;
+}
+
+static const luaL_Reg luaV_module[] = {
+    {"command", luaV_command},
+    {"eval", luaV_eval},
+    {"beep", luaV_beep},
+    {"line", luaV_line},
+    {"buffer", luaV_buffer},
+    {"window", luaV_window},
+    {"open", luaV_open},
+    {"isbuffer", luaV_isbuffer},
+    {"iswindow", luaV_iswindow},
+    {NULL, NULL}
+};
+
+static int luaopen_vim (lua_State *L) {
+    /* set environment */
+    lua_newtable(L);
+    lua_newtable(L);
+    lua_pushliteral(L, "v");
+    lua_setfield(L, -2, "__mode");
+    lua_setmetatable(L, -2);
+    lua_replace(L, LUA_ENVIRONINDEX);
+    /* print */
+    lua_pushcfunction(L, luaV_print);
+    lua_setglobal(L, "print");
+    /* free */
+    lua_pushlightuserdata(L, (void *) LUAVIM_FREE);
+    lua_pushcfunction(L, luaV_free);
+    lua_rawset(L, LUA_REGISTRYINDEX);
+    /* register */
+    luaV_newmetatable(L, LUAVIM_BUFFER);
+    luaL_register(L, NULL, luaV_Buffer_mt);
+    luaV_newmetatable(L, LUAVIM_WINDOW);
+    luaL_register(L, NULL, luaV_Window_mt);
+    luaL_register(L, LUAVIM_NAME, luaV_module);
+    return 0;
+}
+
+static lua_State *luaV_newstate () {
+    lua_State *L = luaL_newstate();
+    const luaL_Reg luaV_core_libs[] = {
+        {"", luaopen_base},
+        {LUA_TABLIBNAME, luaopen_table},
+        {LUA_STRLIBNAME, luaopen_string},
+        {LUA_MATHLIBNAME, luaopen_math},
+        {LUA_OSLIBNAME, luaopen_os}, /* restricted */
+        {LUA_LOADLIBNAME, luaopen_package},
+        {NULL, NULL}
+    };
+    const char *os_funcs[] = {
+        "date", "clock", "time", "difftime", "getenv", NULL
+    };
+    const luaL_Reg *reg = luaV_core_libs;
+    const char **s = os_funcs;
+    /* core libs */
+    for ( ; reg->func; reg++) {
+        lua_pushcfunction(L, reg->func);
+        lua_pushstring(L, reg->name);
+        lua_call(L, 1, 0);
+    }
+    /* restricted os lib */
+    lua_getglobal(L, LUA_OSLIBNAME);
+    lua_newtable(L);
+    for ( ; *s; s++) {
+        lua_getfield(L, -2, *s);
+        lua_setfield(L, -2, *s);
+    }
+    lua_setglobal(L, LUA_OSLIBNAME);
+    lua_pop(L, 1); /* os table */
+    /* vim */
+    lua_pushcfunction(L, luaopen_vim);
+    lua_call(L, 0, 0);
+    return L;
+}
+
+static void luaV_setrange (lua_State *L, int line1, int line2) {
+    lua_getglobal(L, LUAVIM_NAME);
+    lua_pushinteger(L, line1);
+    lua_setfield(L, -2, "firstline");
+    lua_pushinteger(L, line2);
+    lua_setfield(L, -2, "lastline");
+    lua_pop(L, 1); /* vim table */
+}
+
+
+/* =======   Interface   ======= */
+
+static lua_State *L = NULL;
+
+void lua_end () {
+    if (L != NULL) {
+        lua_close(L);
+        L = NULL;
+    }
+}
+
+/* ex commands */
+void ex_lua (exarg_T *eap) {
+    char *script = (char *) script_get(eap, eap->arg);
+    if (L == NULL) L = luaV_newstate(); /* init */
+    if (!eap->skip) {
+        char *s = (script) ? script :  (char *) eap->arg;
+        luaV_setrange(L, eap->line1, eap->line2);
+        if (luaL_loadbuffer(L, s, strlen(s), LUAVIM_CHUNKNAME)
+                || lua_pcall(L, 0, 0, 0)) {
+            EMSG(lua_tostring(L, -1));
+            lua_pop(L, 1);
+        }
+    }
+    if (script != NULL) vim_free(script);
+}
+
+void ex_luado (exarg_T *eap) {
+    linenr_T l;
+    const char *s = (const char *) eap->arg;
+    luaL_Buffer b;
+    size_t len;
+    if (L == NULL) L = luaV_newstate(); /* init */
+    if (u_save(eap->line1 - 1, eap->line2 + 1) == FAIL) {
+        EMSG(_("cannot save undo information"));
+        return;
+    }
+    luaV_setrange(L, eap->line1, eap->line2);
+    luaL_buffinit(L, &b);
+    luaL_addlstring(&b, "return function(line) ", 22); /* header */
+    luaL_addlstring(&b, s, strlen(s));
+    luaL_addlstring(&b, " end", 4); /* footer */
+    luaL_pushresult(&b);
+    s = lua_tolstring(L, -1, &len);
+    if (luaL_loadbuffer(L, s, len, LUAVIM_CHUNKNAME)) {
+        EMSG(lua_tostring(L, -1));
+        lua_pop(L, 2); /* body and msg */
+        return;
+    }
+    lua_call(L, 0, 1);
+    lua_replace(L, -2); /* function -> body */
+    for (l = eap->line1; l <= eap->line2; l++) {
+        lua_pushvalue(L, -1); /* function */
+        lua_pushstring(L, (char *) ml_get(l)); /* current line as arg */
+        if (lua_pcall(L, 1, 1, 0)) {
+            EMSG(lua_tostring(L, -1));
+            lua_pop(L, 1); /* msg */
+            break;
+        }
+        if (lua_isstring(L, -1)) { /* update line? */
+#ifdef HAVE_SANDBOX
+            luaV_checksandbox(L);
+#endif
+            ml_replace(l, (char_u *) lua_tostring(L, -1), TRUE);
+            changed_bytes(l, 0);
+        }
+        lua_pop(L, 1); /* line */
+    }
+    lua_pop(L, 1); /* function */
+    check_cursor();
+    update_screen(NOT_VALID);
+}
+
+void ex_luafile (exarg_T *eap) {
+    if (L == NULL) L = luaV_newstate(); /* init */
+    if (!eap->skip) {
+        luaV_setrange(L, eap->line1, eap->line2);
+        if (luaL_loadfile(L, (char *) eap->arg) || lua_pcall(L, 0, 0, 0)) {
+            EMSG(lua_tostring(L, -1));
+            lua_pop(L, 1);
+        }
+    }
+}
+
+/* buffer */
+void lua_buffer_free (buf_T *buf) {
+    if (L == NULL) L = luaV_newstate(); /* init */
+    luaV_getfield(L, LUAVIM_FREE);
+    lua_pushlightuserdata(L, (void *) buf);
+    lua_call(L, 1, 0);
+}
+
+/* window */
+void lua_window_free (win_T *win) {
+    if (L == NULL) L = luaV_newstate(); /* init */
+    luaV_getfield(L, LUAVIM_FREE);
+    lua_pushlightuserdata(L, (void *) win);
+    lua_call(L, 1, 0);
+}
+
+#endif
diff -Naur vim72-orig/src/main.aap vim72/src/main.aap
--- vim72-orig/src/main.aap	2007-12-07 11:03:31.000000000 -0500
+++ vim72/src/main.aap	2008-09-01 13:13:33.000000000 -0400
@@ -243,13 +243,13 @@
 
 
 PRE_DEFS = -Iproto -I. $DEFS $GUI_DEFS $GUI_IPATH $CPPFLAGS $?(EXTRA_IPATHS)
-POST_DEFS = $X_CFLAGS $MZSCHEME_CFLAGS $PERL_CFLAGS $PYTHON_CFLAGS $TCL_CFLAGS $RUBY_CFLAGS $?(EXTRA_DEFS)
+POST_DEFS = $X_CFLAGS $LUA_CFLAGS $MZSCHEME_CFLAGS $PERL_CFLAGS $PYTHON_CFLAGS $TCL_CFLAGS $RUBY_CFLAGS $?(EXTRA_DEFS)
 CFLAGS = $PRE_DEFS $CONF_CFLAGS $?(PROFILE_CFLAGS) $POST_DEFS
 CPPFLAGS =
 
 ALL_LIB_DIRS = $GUI_LIBS_DIR $X_LIBS_DIR
 LDFLAGS = $ALL_LIB_DIRS $CONF_LDFLAGS
-LIBS = $GUI_LIBS1 $GUI_X_LIBS $GUI_LIBS2 $X_PRE_LIBS $X_LIBS $X_EXTRA_LIBS $CONF_LIBS $?(EXTRA_LIBS) $MZSCHEME_LIBS $PERL_LIBS $PYTHON_LIBS $TCL_LIBS $RUBY_LIBS $?(PROFILE_LIBS)
+LIBS = $GUI_LIBS1 $GUI_X_LIBS $GUI_LIBS2 $X_PRE_LIBS $X_LIBS $X_EXTRA_LIBS $CONF_LIBS $?(EXTRA_LIBS) $LUA_LIBS $MZSCHEME_LIBS $PERL_LIBS $PYTHON_LIBS $TCL_LIBS $RUBY_LIBS $?(PROFILE_LIBS)
 
 Target = $VIMNAME
 
@@ -337,6 +337,7 @@
         $OS_EXTRA_SRC
         $GUI_SRC
         $HANGULIN_SRC
+        $LUA_SRC
         $MZSCHEME_SRC
         $PERL_SRC
         $NETBEANS_SRC
diff -Naur vim72-orig/src/main.c vim72/src/main.c
--- vim72-orig/src/main.c	2008-07-24 04:40:56.000000000 -0400
+++ vim72/src/main.c	2008-08-30 10:44:24.000000000 -0400
@@ -1308,6 +1308,9 @@
 	windgoto((int)Rows - 1, 0);
 #endif
 
+#ifdef FEAT_LUA
+    lua_end();
+#endif
 #ifdef FEAT_MZSCHEME
     mzscheme_end();
 #endif
diff -Naur vim72-orig/src/Makefile vim72/src/Makefile
--- vim72-orig/src/Makefile	2008-08-09 10:50:04.000000000 -0400
+++ vim72/src/Makefile	2008-08-30 18:55:28.000000000 -0400
@@ -397,6 +397,13 @@
 # Uncomment this when you want to include the Ruby interface.
 #CONF_OPT_RUBY = --enable-rubyinterp
 
+# LUA
+# Uncomment this when you want to include the Lua interface.
+#CONF_OPT_LUA = --enable-luainterp
+# Lua installation dir (also in LUA_PREFIX environment variable)
+#CONF_OPT_LUA_PREFIX = --with-lua-prefix=/usr/local
+#CONF_OPT_LUA_PREFIX = --with-lua-prefix=/usr
+
 # MZSCHEME
 # Uncomment this when you want to include the MzScheme interface.
 #CONF_OPT_MZSCHEME = --enable-mzschemeinterp
@@ -1291,7 +1298,7 @@
 .SUFFIXES: .c .o .pro
 
 PRE_DEFS = -Iproto $(DEFS) $(GUI_DEFS) $(GUI_IPATH) $(CPPFLAGS) $(EXTRA_IPATHS)
-POST_DEFS = $(X_CFLAGS) $(MZSCHEME_CFLAGS) $(PERL_CFLAGS) $(PYTHON_CFLAGS) $(TCL_CFLAGS) $(RUBY_CFLAGS) $(EXTRA_DEFS)
+POST_DEFS = $(X_CFLAGS) $(LUA_CFLAGS) $(MZSCHEME_CFLAGS) $(PERL_CFLAGS) $(PYTHON_CFLAGS) $(TCL_CFLAGS) $(RUBY_CFLAGS) $(EXTRA_DEFS)
 
 ALL_CFLAGS = $(PRE_DEFS) $(CFLAGS) $(PROFILE_CFLAGS) $(POST_DEFS)
 
@@ -1311,7 +1318,7 @@
 PFLAGS = $(PROTO_FLAGS) -DPROTO $(LINT_CFLAGS)
 
 ALL_LIB_DIRS = $(GUI_LIBS_DIR) $(X_LIBS_DIR)
-ALL_LIBS = $(GUI_LIBS1) $(GUI_X_LIBS) $(GUI_LIBS2) $(X_PRE_LIBS) $(X_LIBS) $(X_EXTRA_LIBS) $(LIBS) $(EXTRA_LIBS) $(MZSCHEME_LIBS) $(PERL_LIBS) $(PYTHON_LIBS) $(TCL_LIBS) $(RUBY_LIBS) $(PROFILE_LIBS)
+ALL_LIBS = $(GUI_LIBS1) $(GUI_X_LIBS) $(GUI_LIBS2) $(X_PRE_LIBS) $(X_LIBS) $(X_EXTRA_LIBS) $(LIBS) $(EXTRA_LIBS) $(LUA_LIBS) $(MZSCHEME_LIBS) $(PERL_LIBS) $(PYTHON_LIBS) $(TCL_LIBS) $(RUBY_LIBS) $(PROFILE_LIBS)
 
 # abbreviations
 DEST_BIN = $(DESTDIR)$(BINDIR)
@@ -1412,13 +1419,13 @@
 	window.c \
 	$(OS_EXTRA_SRC)
 
-SRC =	$(BASIC_SRC) $(GUI_SRC) $(HANGULIN_SRC) $(MZSCHEME_SRC) \
+SRC =	$(BASIC_SRC) $(GUI_SRC) $(HANGULIN_SRC) $(LUA_SRC) $(MZSCHEME_SRC) \
 	$(PERL_SRC) $(PYTHON_SRC) $(TCL_SRC) $(RUBY_SRC) \
 	$(SNIFF_SRC) $(WORKSHOP_SRC) $(WSDEBUG_SRC)
 
 TAGS_SRC = *.c *.cpp if_perl.xs
 
-EXTRA_SRC = hangulin.c if_mzsch.c auto/if_perl.c if_perlsfio.c \
+EXTRA_SRC = hangulin.c if_lua.c if_mzsch.c auto/if_perl.c if_perlsfio.c \
 	    if_python.c if_tcl.c if_ruby.c if_sniff.c gui_beval.c \
 	    workshop.c wsdebug.c integration.c netbeans.c
 
@@ -1482,6 +1489,7 @@
 	objects/undo.o \
 	objects/window.o \
 	$(GUI_OBJ) \
+	$(LUA_OBJ) \
 	$(MZSCHEME_OBJ) \
 	$(PERL_OBJ) \
 	$(PYTHON_OBJ) \
@@ -1580,6 +1588,7 @@
 		$(CONF_OPT_SNIFF) $(CONF_OPT_FEAT) $(CONF_TERM_LIB) \
 		$(CONF_OPT_COMPBY) $(CONF_OPT_ACL)  $(CONF_OPT_NETBEANS) \
 		$(CONF_ARGS) $(CONF_OPT_MZSCHEME) $(CONF_OPT_PLTHOME) \
+		$(CONF_OPT_LUA) $(CONF_OPT_LUA_PREFIX) \
 		$(CONF_OPT_SYSMOUSE)
 
 # Use "make reconfig" to rerun configure without cached values.
@@ -2420,6 +2429,9 @@
 objects/if_xcmdsrv.o: if_xcmdsrv.c
 	$(CCC) -o $@ if_xcmdsrv.c
 
+objects/if_lua.o: if_lua.c
+	$(CCC) -o $@ if_lua.c
+
 objects/if_mzsch.o: if_mzsch.c
 	$(CCC) -o $@ if_mzsch.c
 
@@ -2912,6 +2924,10 @@
   auto/osdef.h ascii.h keymap.h term.h macros.h option.h structs.h \
   regexp.h gui.h gui_beval.h proto/gui_beval.pro ex_cmds.h proto.h \
   globals.h farsi.h arabic.h
+objects/if_lua.o: if_lua.c vim.h auto/config.h feature.h os_unix.h \
+  auto/osdef.h ascii.h keymap.h term.h macros.h option.h structs.h \
+  regexp.h gui.h gui_beval.h proto/gui_beval.pro ex_cmds.h proto.h \
+  globals.h farsi.h arabic.h
 objects/if_mzsch.o: if_mzsch.c vim.h auto/config.h feature.h os_unix.h \
   auto/osdef.h ascii.h keymap.h term.h macros.h option.h structs.h \
   regexp.h gui.h gui_beval.h proto/gui_beval.pro ex_cmds.h proto.h \
diff -Naur vim72-orig/src/proto/if_lua.pro vim72/src/proto/if_lua.pro
--- vim72-orig/src/proto/if_lua.pro	1969-12-31 19:00:00.000000000 -0500
+++ vim72/src/proto/if_lua.pro	2008-08-30 23:13:35.000000000 -0400
@@ -0,0 +1,8 @@
+/* if_lua.c */
+void lua_end __ARGS((void));
+void ex_lua __ARGS((exarg_T *eap));
+void ex_luado __ARGS((exarg_T *eap));
+void ex_luafile __ARGS((exarg_T *eap));
+void lua_buffer_free __ARGS((buf_T *buf));
+void lua_window_free __ARGS((win_T *win));
+/* vim: set ft=c : */
diff -Naur vim72-orig/src/proto.h vim72/src/proto.h
--- vim72-orig/src/proto.h	2006-03-24 12:15:37.000000000 -0500
+++ vim72/src/proto.h	2008-08-30 10:45:50.000000000 -0400
@@ -159,6 +159,10 @@
 # include "version.pro"
 # include "window.pro"
 
+# ifdef FEAT_LUA
+#  include "if_lua.pro"
+# endif
+
 # ifdef FEAT_MZSCHEME
 #  include "if_mzsch.pro"
 # endif
diff -Naur vim72-orig/src/version.c vim72/src/version.c
--- vim72-orig/src/version.c	2008-08-09 10:24:52.000000000 -0400
+++ vim72/src/version.c	2008-08-30 10:46:50.000000000 -0400
@@ -304,6 +304,15 @@
 #else
 	"-localmap",
 #endif
+#ifdef FEAT_LUA
+# ifdef DYNAMIC_LUA
+	"+lua/dyn",
+# else
+	"+lua",
+# endif
+#else
+	"-lua",
+#endif
 #ifdef FEAT_MENU
 	"+menu",
 #else
diff -Naur vim72-orig/src/window.c vim72/src/window.c
--- vim72-orig/src/window.c	2008-08-06 07:00:30.000000000 -0400
+++ vim72/src/window.c	2008-08-30 10:47:33.000000000 -0400
@@ -4243,6 +4243,10 @@
     block_autocmds();
 #endif
 
+#ifdef FEAT_LUA
+    lua_window_free(wp);
+#endif
+
 #ifdef FEAT_MZSCHEME
     mzscheme_window_free(wp);
 #endif
